



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
echo *** Test Cases *** >> test-6.1.1.robot
IF %ERRORLEVEL% NEQ 0 exit /B 1
echo|(set /p="Hello World    Log   %%{PKG_NAME} %%{PKG_VERSION}" & echo.) >> test-6.1.1.robot
IF %ERRORLEVEL% NEQ 0 exit /B 1
robot test-6.1.1.robot
IF %ERRORLEVEL% NEQ 0 exit /B 1
rebot output.xml
IF %ERRORLEVEL% NEQ 0 exit /B 1
libdoc test-6.1.1.robot test.html
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
